@@ -1,22 +1,35 @@
-#!/usr/bin/perl
+package Data::Visitor; # git description: Data-Visitor-0.27-23-g8b10b8a
+use Moose;
+# ABSTRACT: Visitor style traversal of Perl data structures
 
-package Data::Visitor;
-use base qw/Class::Accessor/;
-
-use strict;
-use warnings;
-
-use Scalar::Util qw/blessed refaddr reftype/;
+our $VERSION = '0.31';
+use Scalar::Util qw/blessed refaddr reftype weaken isweak/;
 use overload ();
 use Symbol ();
 
 use Tie::ToObject;
 
-__PACKAGE__->mk_accessors(qw(tied_as_objects));
+no warnings 'recursion';
 
-# the double not works makes this no longer undef, so exempt from useless constant warnings in older perls
+use namespace::clean -except => 'meta';
+
+# the double not makes this no longer undef, so exempt from useless constant warnings in older perls
 use constant DEBUG => not not our $DEBUG || $ENV{DATA_VISITOR_DEBUG};
 
+use constant HAS_DATA_ALIAS => eval { +require Data::Alias; 1 };
+
+has tied_as_objects => (
+	isa => "Bool",
+	is  => "rw",
+);
+
+# currently broken
+has weaken => (
+	isa => "Bool",
+	is  => "rw",
+	default => 0,
+);
+
 sub trace {
 	my ( $self, $category, @msg ) = @_;
 
@@ -35,25 +48,37 @@ sub _print_trace {
 	warn "@msg\n";
 }
 
-our $VERSION = "0.17";
-
 sub visit {
-	my ( $self, $data ) = @_;
+	my $self = shift;
 
 	local $self->{depth} = (($self->{depth}||0) + 1) if DEBUG;
-	$self->trace( flow => visit => $data ) if DEBUG;
-
 	my $seen_hash = local $self->{_seen} = ($self->{_seen} || {}); # delete it after we're done with the whole visit
-	if ( ref $data ) { # only references need recursion checks
-		if ( exists $seen_hash->{ refaddr($data) } ) {
-			$self->trace( mapping => found_mapping => from => $data, to => $seen_hash->{ refaddr($data) } ) if DEBUG;
-			return $self->visit_seen( $data, $seen_hash->{refaddr($data)} );
+
+	my @ret;
+
+	foreach my $data ( @_ ) {
+		$self->trace( flow => visit => $data ) if DEBUG;
+
+		if ( my $refaddr = ref($data) && refaddr($data) ) { # only references need recursion checks
+			$seen_hash->{weak} ||= isweak($data) if $self->weaken;
+
+			if ( exists $seen_hash->{$refaddr} ) {
+				$self->trace( mapping => found_mapping => from => $data, to => $seen_hash->{$refaddr} ) if DEBUG;
+				push @ret, $self->visit_seen( $data, $seen_hash->{$refaddr} );
+				next;
+			} else {
+				$self->trace( mapping => no_mapping => $data ) if DEBUG;
+			}
+		}
+
+		if ( defined wantarray ) {
+			push @ret, scalar($self->visit_no_rec_check($data));
 		} else {
-			$self->trace( mapping => no_mapping => $data ) if DEBUG;
+			$self->visit_no_rec_check($data);
 		}
 	}
 
-	return $self->visit_no_rec_check( $data );
+	return ( @_ == 1 ? $ret[0] : @ret );
 }
 
 sub visit_seen {
@@ -77,18 +102,25 @@ sub visit_no_rec_check {
 	my ( $self, $data ) = @_;
 
 	if ( blessed($data) ) {
-		return $self->visit_object($data);
+		return $self->visit_object($_[1]);
 	} elsif ( ref $data ) {
-		return $self->visit_ref($data);
+		return $self->visit_ref($_[1]);
 	}
 
-	return $self->visit_value($data);
+	return $self->visit_value($_[1]);
 }
 
 sub visit_object {
 	my ( $self, $object ) = @_;
 	$self->trace( flow => visit_object => $object ) if DEBUG;
-	return $self->_register_mapping( $object, $self->visit_value($object) );
+
+	if ( not defined wantarray ) {
+		$self->_register_mapping( $object, $object );
+		$self->visit_value($_[1]);
+		return;
+	} else {
+		return $self->_register_mapping( $object, $self->visit_value($_[1]) );
+	}
 }
 
 sub visit_ref {
@@ -104,7 +136,7 @@ sub visit_ref {
 
 	my $method = $self->can(lc "visit_$reftype") || "visit_value";
 
-	return $self->$method($data);
+	return $self->$method($_[1]);
 }
 
 sub visit_value {
@@ -118,30 +150,48 @@ sub visit_hash {
 
 	local $self->{depth} = (($self->{depth}||0) + 1) if DEBUG;
 
-	if ( not defined wantarray ) {
-		$self->_register_mapping( $hash, $hash );
-
-		my $tied = tied(%$hash);
-		if ( ref($tied) and $self->tied_as_objects ) {
-			$self->visit_tied($tied, $hash);
-		} else {
-			$self->visit_hash_entries($hash);
-		}
-
-		return;
+	if ( defined(tied(%$hash)) and $self->tied_as_objects ) {
+		return $self->visit_tied_hash(tied(%$hash), $_[1]);
 	} else {
+		return $self->visit_normal_hash($_[1]);
+	}
+}
+
+sub visit_normal_hash {
+	my ( $self, $hash ) = @_;
+
+	if ( defined wantarray ) {
 		my $new_hash = {};
 		$self->_register_mapping( $hash, $new_hash );
 
-		my $tied = tied(%$hash);
-		if ( ref($tied) and $self->tied_as_objects and blessed(my $new_tied = $self->visit_tied($tied, $hash)) ) {
+		%$new_hash = $self->visit_hash_entries($_[1]);
+
+		return $self->retain_magic( $_[1], $new_hash );
+	} else {
+		$self->_register_mapping($hash, $hash);
+		$self->visit_hash_entries($_[1]);
+		return;
+	}
+}
+
+sub visit_tied_hash {
+	my ( $self, $tied, $hash ) = @_;
+
+	if ( defined wantarray ) {
+		my $new_hash = {};
+		$self->_register_mapping( $hash, $new_hash );
+
+		if ( blessed(my $new_tied = $self->visit_tied($_[1], $_[2])) ) {
 			$self->trace( data => tying => var => $new_hash, to => $new_tied ) if DEBUG;
 			tie %$new_hash, 'Tie::ToObject', $new_tied;
+			return $self->retain_magic($_[2], $new_hash);
 		} else {
-			%$new_hash = $self->visit_hash_entries($hash);
+			return $self->visit_normal_hash($_[2]);
 		}
-
-		return $self->retain_magic( $hash, $new_hash );
+	} else {
+		$self->_register_mapping($hash, $hash);
+		$self->visit_tied($_[1], $_[2]);
+		return;
 	}
 }
 
@@ -162,11 +212,11 @@ sub visit_hash_entry {
 
 	if ( not defined wantarray ) {
 		$self->visit_hash_key($key,$value,$hash);
-		$self->visit_hash_value($_[2],$key,$hash); # retain aliasing semantics
+		$self->visit_hash_value($_[2],$key,$hash);
 	} else {
 		return (
 			$self->visit_hash_key($key,$value,$hash),
-			$self->visit_hash_value($_[2],$key,$hash) # retain aliasing semantics
+			$self->visit_hash_value($_[2],$key,$hash),
 		);
 	}
 }
@@ -178,36 +228,56 @@ sub visit_hash_key {
 
 sub visit_hash_value {
 	my ( $self, $value, $key, $hash ) = @_;
-	$self->visit($_[1]); # retain it's aliasing semantics
+	$self->visit($_[1]);
 }
 
 sub visit_array {
 	my ( $self, $array ) = @_;
 
-	if ( not defined wantarray ) {
-		$self->_register_mapping( $array, $array );
-
-		my $tied = tied(@$array);
-		if ( ref($tied) and $self->tied_as_objects ) {
-			$self->visit_tied($tied, $array);
-		} else {
-			$self->visit_array_entries($array);
-		}
-
-		return;
+	if ( defined(tied(@$array)) and $self->tied_as_objects ) {
+		return $self->visit_tied_array(tied(@$array), $_[1]);
 	} else {
+		return $self->visit_normal_array($_[1]);
+	}
+}
+
+sub visit_normal_array {
+	my ( $self, $array ) = @_;
+
+	if ( defined wantarray ) {
 		my $new_array = [];
 		$self->_register_mapping( $array, $new_array );
 
-		my $tied = tied(@$array);
-		if ( ref($tied) and $self->tied_as_objects and blessed(my $new_tied = $self->visit_tied($tied, $array)) ) {
-			$self->trace( data => tying => var => $new_array, to => $new_tied ) if DEBUG;
-			tie @$new_array, 'Data::Visitor::TieToObject', $new_tied;
-		} else {
-			@$new_array = $self->visit_array_entries($array);
-		}
+		@$new_array = $self->visit_array_entries($_[1]);
 
-		return $self->retain_magic( $array, $new_array );
+		return $self->retain_magic( $_[1], $new_array );
+	} else {
+		$self->_register_mapping( $array, $array );
+		$self->visit_array_entries($_[1]);
+
+		return;
+	}
+}
+
+sub visit_tied_array {
+	my ( $self, $tied, $array ) = @_;
+
+	if ( defined wantarray ) {
+		my $new_array = [];
+		$self->_register_mapping( $array, $new_array );
+
+		if ( blessed(my $new_tied = $self->visit_tied($_[1], $_[2])) ) {
+			$self->trace( data => tying => var => $new_array, to => $new_tied ) if DEBUG;
+			tie @$new_array, 'Tie::ToObject', $new_tied;
+			return $self->retain_magic($_[2], $new_array);
+		} else {
+			return $self->visit_normal_array($_[2]);
+		}
+	} else {
+		$self->_register_mapping( $array, $array );
+		$self->visit_tied($_[1], $_[2]);
+
+		return;
 	}
 }
 
@@ -229,67 +299,103 @@ sub visit_array_entry {
 sub visit_scalar {
 	my ( $self, $scalar ) = @_;
 
-	if ( not defined wantarray ) {
-		$self->_register_mapping( $scalar, $scalar );
-
-		my $tied = tied($$scalar);
-		if ( ref($tied) and $self->tied_as_objects ) {
-			$self->visit_tied($tied, $scalar);
-		} else {
-			$self->visit($$scalar);
-		}
-
-		return;
+	if ( defined(tied($$scalar)) and $self->tied_as_objects ) {
+		return $self->visit_tied_scalar(tied($$scalar), $_[1]);
 	} else {
+		return $self->visit_normal_scalar($_[1]);
+	}
+}
+
+sub visit_normal_scalar {
+	my ( $self, $scalar ) = @_;
+
+	if ( defined wantarray ) {
 		my $new_scalar;
 		$self->_register_mapping( $scalar, \$new_scalar );
 
-		my $tied = tied($$scalar);
-		if ( ref($tied) and $self->tied_as_objects and blessed(my $new_tied = $self->visit_tied($tied, $scalar)) ) {
-			$self->trace( data => tying => var => $new_scalar, to => $new_tied ) if DEBUG;
-			tie $new_scalar, 'Data::Visitor::TieToObject', $new_tied;
-		} else {
-			$new_scalar = $self->visit( $$scalar );
-		}
+		$new_scalar = $self->visit( $$scalar );
 
-		return $self->retain_magic( $scalar, \$new_scalar );
+		return $self->retain_magic($_[1], \$new_scalar);
+	} else {
+		$self->_register_mapping( $scalar, $scalar );
+		$self->visit( $$scalar );
+		return;
+	}
+
+}
+
+sub visit_tied_scalar {
+	my ( $self, $tied, $scalar ) = @_;
+
+	if ( defined wantarray ) {
+		my $new_scalar;
+		$self->_register_mapping( $scalar, \$new_scalar );
+
+		if ( blessed(my $new_tied = $self->visit_tied($_[1], $_[2])) ) {
+			$self->trace( data => tying => var => $new_scalar, to => $new_tied ) if DEBUG;
+			tie $new_scalar, 'Tie::ToObject', $new_tied;
+			return $self->retain_magic($_[2], \$new_scalar);
+		} else {
+			return $self->visit_normal_scalar($_[2]);
+		}
+	} else {
+		$self->_register_mapping( $scalar, $scalar );
+		$self->visit_tied($_[1], $_[2]);
+		return;
 	}
 }
 
 sub visit_code {
 	my ( $self, $code ) = @_;
-	$self->visit_value($code);
+	$self->visit_value($_[1]);
 }
 
 sub visit_glob {
 	my ( $self, $glob ) = @_;
 
-	if ( not defined wantarray ) {
-		$self->_register_mapping( $glob, $glob );
-
-		my $tied = tied(*$glob);
-		if ( ref($tied) and $self->tied_as_objects ) {
-			$self->visit_tied($tied, $glob);
-		} else {
-			$self->visit( *$glob{$_} || next ) for qw/SCALAR ARRAY HASH/;
-		}
-
-		return;
+	if ( defined(tied(*$glob)) and $self->tied_as_objects ) {
+		return $self->visit_tied_glob(tied(*$glob), $_[1]);
 	} else {
-		my $new_glob = Symbol::gensym();
+		return $self->visit_normal_glob($_[1]);
+	}
+}
 
+sub visit_normal_glob {
+	my ( $self, $glob ) = @_;
+
+	if ( defined wantarray ) {
+		my $new_glob = Symbol::gensym();
 		$self->_register_mapping( $glob, $new_glob );
 
-		my $tied = tied(*$glob);
-		if ( ref($tied) and $self->tied_as_objects and blessed(my $new_tied = $self->visit_tied($tied, $glob)) ) {
-			$self->trace( data => tying => var => $new_glob, to => $new_tied ) if DEBUG;
-			tie *$new_glob, 'Data::Visitor::TieToObject', $new_tied;
-		} else {
-			no warnings 'misc'; # Undefined value assigned to typeglob
-			*$new_glob = $self->visit( *$glob{$_} || next ) for qw/SCALAR ARRAY HASH/;
-		}
+		no warnings 'misc'; # Undefined value assigned to typeglob
+		*$new_glob = $self->visit( *$glob{$_} || next ) for qw/SCALAR ARRAY HASH/;
 
-		return $self->retain_magic( $glob, $new_glob );
+		return $self->retain_magic($_[1], $new_glob);
+	} else {
+		$self->_register_mapping( $glob, $glob );
+		$self->visit( *$glob{$_} || next ) for qw/SCALAR ARRAY HASH/;
+		return;
+	}
+}
+
+sub visit_tied_glob {
+	my ( $self, $tied, $glob ) = @_;
+
+	if ( defined wantarray ) {
+		my $new_glob = Symbol::gensym();
+		$self->_register_mapping( $glob, \$new_glob );
+
+		if ( blessed(my $new_tied = $self->visit_tied($_[1], $_[2])) ) {
+			$self->trace( data => tying => var => $new_glob, to => $new_tied ) if DEBUG;
+			tie *$new_glob, 'Tie::ToObject', $new_tied;
+			return $self->retain_magic($_[2], $new_glob);
+		} else {
+			return $self->visit_normal_glob($_[2]);
+		}
+	} else {
+		$self->_register_mapping( $glob, $glob );
+		$self->visit_tied($_[1], $_[2]);
+		return;
 	}
 }
 
@@ -301,6 +407,35 @@ sub retain_magic {
 		bless $new, ref $proto;
 	}
 
+	my $seen_hash = $self->{_seen};
+	if ( $seen_hash->{weak} ) {
+		#if ("$]" >= '5.022') {
+		#  TODO: Data::Alias does not work on recent perls, but there is built-in aliasing support now.
+		#  e.g. see what Var::Pairs 0.003004 did.
+		#}
+		if (HAS_DATA_ALIAS) {
+			my @weak_refs;
+			foreach my $value ( Data::Alias::deref($proto) ) {
+				if ( ref $value and isweak($value) ) {
+					push @weak_refs, refaddr $value;
+				}
+			}
+
+			if ( @weak_refs ) {
+				my %targets = map { refaddr($_) => 1 } @{ $self->{_seen} }{@weak_refs};
+				foreach my $value ( Data::Alias::deref($new) ) {
+					if ( ref $value and $targets{refaddr($value)}) {
+						push @{ $seen_hash->{weakened} ||= [] }, $value; # keep a ref around
+						weaken($value);
+					}
+				}
+			}
+		}
+		else {
+			die "Found a weak reference, but Data::Alias is not installed. You must install Data::Alias in order for this to work.";
+		}
+	}
+
 	# FIXME real magic, too
 
 	return $new;
@@ -309,34 +444,48 @@ sub retain_magic {
 sub visit_tied {
 	my ( $self, $tied, $var ) = @_;
 	$self->trace( flow => visit_tied => $tied ) if DEBUG;
-	$self->visit($tied); # as an object eventually
+	$self->visit($_[1]); # as an object eventually
 }
 
+__PACKAGE__->meta->make_immutable if __PACKAGE__->meta->can("make_immutable");
+
 __PACKAGE__;
 
 __END__
 
 =pod
 
+=encoding UTF-8
+
 =head1 NAME
 
 Data::Visitor - Visitor style traversal of Perl data structures
 
+=head1 VERSION
+
+version 0.31
+
 =head1 SYNOPSIS
 
 	# NOTE
 	# You probably want to use Data::Visitor::Callback for trivial things
 
 	package FooCounter;
-	use base qw/Data::Visitor/;
+	use Moose;
 
-	BEGIN { __PACKAGE__->mk_accessors( "number_of_foos" ) };
+	extends qw(Data::Visitor);
+
+	has number_of_foos => (
+		isa => "Int",
+		is  => "rw",
+		default => 0,
+	);
 
 	sub visit_value {
 		my ( $self, $data ) = @_;
 
 		if ( defined $data and $data eq "foo" ) {
-			$self->number_of_foos( ($self->number_of_foos || 0) + 1 );
+			$self->number_of_foos( $self->number_of_foos + 1 );
 		}
 
 		return $data;
@@ -372,7 +521,7 @@ driven subclass, L<Data::Visitor::Callback>.
 
 =item visit $data
 
-This method takes any Perl value as it's only argument, and dispatches to the
+This method takes any Perl value as its only argument, and dispatches to the
 various other visiting methods using C<visit_no_rec_check>, based on the data's
 type.
 
@@ -381,7 +530,7 @@ called.
 
 =item visit_seen $data, $first_result
 
-When an already seen value is encountered again it's typically replaced with
+When an already seen value is encountered again, it is typically replaced with
 the result of the first visitation of that value. The value and the result of
 the first visitation are passed as arguments.
 
@@ -475,7 +624,6 @@ copying should be.
 
 Called if the C<DEBUG> constant is set with a trace message.
 
-
 =back
 
 =head1 RETURN VALUE
@@ -484,13 +632,12 @@ This object can be used as an C<fmap> of sorts - providing an ad-hoc functor
 interface for Perl data structures.
 
 In void context this functionality is ignored, but in any other context the
-default methods will all try to return a value of similar structure, with it's
+default methods will all try to return a value of similar structure, with its
 children also fmapped.
 
 =head1 SUBCLASSING
 
-Create instance data using the L<Class::Accessor> interface. L<Data::Visitor>
-inherits L<Class::Accessor> to get a sane C<new>.
+Data::Visitor is a L<Moose> class, so it should be subclassed using Moose.
 
 Then override the callback methods in any way you like. To retain visitor
 behavior, make sure to retain the functionality of C<visit_array> and
@@ -519,16 +666,68 @@ L<http://en.wikipedia.org/wiki/Visitor_pattern>,
 L<http://www.ninebynine.org/Software/Learning-Haskell-Notes.html#functors>,
 L<http://en.wikipedia.org/wiki/Functor>
 
-=head1 AUTHOR
+=for Pod::Coverage HAS_DATA_ALIAS
+visit_normal_array
+visit_normal_glob
+visit_normal_hash
+visit_normal_scalar
+visit_tied_array
+visit_tied_glob
+visit_tied_hash
+visit_tied_scalar
+
+=head1 SUPPORT
+
+Bugs may be submitted through L<the RT bug tracker|https://rt.cpan.org/Public/Dist/Display.html?Name=Data-Visitor>
+(or L<bug-Data-Visitor@rt.cpan.org|mailto:bug-Data-Visitor@rt.cpan.org>).
+
+=head1 AUTHORS
+
+=over 4
+
+=item *
 
 Yuval Kogman <nothingmuch@woobling.org>
 
-=head1 COPYRIGHT & LICENSE
+=item *
 
-	Copyright (c) 2006-2008 Yuval Kogman. All rights reserved
-	This program is free software; you can redistribute
-	it and/or modify it under the same terms as Perl itself.
+Marcel Grünauer <marcel@cpan.org>
+
+=back
+
+=head1 CONTRIBUTORS
+
+=for stopwords Jesse Luehrs Florian Ragwitz Karen Etheridge David Steinbrunner Robin Smidsrød
+
+=over 4
+
+=item *
+
+Jesse Luehrs <doy@tozt.net>
+
+=item *
+
+Florian Ragwitz <rafl@debian.org>
+
+=item *
+
+Karen Etheridge <ether@cpan.org>
+
+=item *
+
+David Steinbrunner <dsteinbrunner@pobox.com>
+
+=item *
+
+Robin Smidsrød <robin@smidsrod.no>
+
+=back
+
+=head1 COPYRIGHT AND LICENCE
+
+This software is copyright (c) 2020 by Yuval Kogman.
+
+This is free software; you can redistribute it and/or modify it under
+the same terms as the Perl 5 programming language system itself.
 
 =cut
-
-
